from types import FunctionType
from fastmcp.prompts import Prompt

from dlt_mcp._prompts.infer_table_reference import infer_table_reference


__all__ = [
    "PROMPTS_REGISTRY",
]


PROMPTS_REGISTRY: dict[str, Prompt] = {}


def register_prompt(fn: FunctionType) -> FunctionType:
    global PROMPTS_REGISTRY
    PROMPTS_REGISTRY[fn.__name__] = Prompt.from_function(fn)
    return fn


register_prompt(infer_table_reference)



from typing import Literal

import lancedb

from dlt_mcp._utilities.ingestion import (
    db_con,
    DLT_VERSION,
    DLT_DOCS_CHUNKS_TABLE_NAME,
    DLT_CODE_CHUNKS_TABLE_NAME,
    _maybe_ingest_docs_and_code,
)

LOCAL_DATA_IS_AVAILABLE = False
"""Global flag indicating the local LanceDB database powering search tools
is populated. This is asserted on the first search tool call.

We chose to conduct on first tool call instead of module init or server startup
to avoid blocking the server or slowing down operations unrelated to search. 
"""


def _ensure_docs_and_code_ingestion():
    global LOCAL_DATA_IS_AVAILABLE
    if not LOCAL_DATA_IS_AVAILABLE:
        _maybe_ingest_docs_and_code(DLT_VERSION)
        LOCAL_DATA_IS_AVAILABLE = True


def _retrieve_docs(
    table: lancedb.Table, query: str, query_type: Literal["fts", "vector", "hybrid"]
) -> list[dict]:
    retrieval_query = (
        table.search(query, query_type=query_type)
        .select(["text", "file_path"])
        .limit(3)
    )
    results = retrieval_query.to_list()
    return results


# TODO improve docstring to instruct with `mode` to use
# TODO maybe it doesn't need `vector` and
def search_docs(
    query: str, mode: Literal["hybrid", "full_text", "vector"] = "full_text"
) -> list[dict] | str:
    """Search over the `dlt` documentation. Use it to verify if a feature
    exists, answer general questions, or identify recommended patterns.
    """
    # TODO find a more elegant mechanism
    _ensure_docs_and_code_ingestion()

    query_type: Literal["fts", "vector", "hybrid"] = (
        "fts" if mode == "full_text" else mode
    )

    db = db_con(dlt_version=DLT_VERSION)
    table = db.open_table(DLT_DOCS_CHUNKS_TABLE_NAME)
    results = _retrieve_docs(table=table, query=query, query_type=query_type)
    if not results:
        return (
            f"No docs result found for `{query=}` and `{query_type=}`."
            " Consider using a different `query_type` or using a more appropriate `query`"
            " for this query type."
        )
    else:
        return results


def _retrieve_code(
    table: lancedb.Table,
    query: str,
    file_path: str | None,
) -> list[dict]:
    retrieval_query = (
        table.search(query, query_type="fts").select(["text", "file_path"]).limit(3)
    )
    if file_path:
        retrieval_query = retrieval_query.where(
            f"file_path = '{file_path}'", prefilter=True
        )
    results = retrieval_query.to_list()
    return results


# The source code search could degrade performance given the majority of
# code is internal and not public-facing APIs. It could help debug though.
def search_code(query: str, file_path: str | None = None) -> list[dict] | str:
    # TODO find a more elegant mechanism
    _ensure_docs_and_code_ingestion()

    db = db_con(dlt_version=DLT_VERSION)
    table = db.open_table(DLT_CODE_CHUNKS_TABLE_NAME)
    results = _retrieve_code(table=table, query=query, file_path=file_path)
    if not results:
        return (
            f"No code result found for `{query=}` and `{file_path=}`."
            " Consider using a different `file_path` filter or modifying the `query`."
        )
    else:
        return results



from typing import Optional
from dlt.common.schema.typing import TTableReference


def infer_table_reference(pipeline_name: Optional[str] = None):
    """Generates guidelines to infer table references for a given pipeline"""
    table_reference_documentation = _get_table_reference_documentation()

    prompt = (
        "You are an helpful assistant to data architect and data engineers using DLT tasked with analyzing table relationships within a dlt pipeline.\n"
        "## Workflow Steps \n"
    )

    if pipeline_name is None:
        prompt += (
            "- **Pipeline Discovery**:\n"
            "   - First, list all available pipelines using `list_pipelines()`\n"
            "   - Ask the user which pipeline they want to investigate \n"
            "   - wait until the user has provided a valid pipeline name"
        )
    prompt += (
        "-  **Data Exploration**:\n"
        "   - Ask the User if they want to explore specific tables or all of them \n"
        "   - Get schema details for each table using `get_table_schema(pipeline_name, table_name)`\n"
        "   - Show schema using mermaid, this will help make conversation more easier and save the world from bad data schemas \n"
        "- **Relationship Analysis**:\n"
        "   - Look for common column names across tables (e.g., 'user_id', 'customer_id')\n"
        "   - Identify foreign key patterns (e.g., 'parent_id', 'foreign_key')\n"
        "   - Check for date/time columns that might indicate relationships\n"
        "   - Examine table descriptions for hints about relationships\n"
        "   - Look for auto-incrementing IDs that might reference other tables\n"
        "   - Generate mermaid to showcase the relationships. This too helps save the world from bad data\n"
        "- **Validate Relationships**:\n"
        "   - Suggest ways to confirm these relationships (e.g., sample data inspection, referential integrity checks)\n"
        "   - You can execute these validations using execute_sql_query tool \n"
        "- **Generating Table References**: \n"
        f"  - {table_reference_documentation} \n"
        "   - To maintain the information across codebase it's important to generate the table reference in the above format \n"
        "## Tips:\n"
        "- Use only the tools available to go through the process \n"
        "- keep explanations small and to the point until asked for more details \n"
        "- Think before providing reasoning about the relationships and one by one confirm each of them with the user \n"
        "- NO NEED TO EXPLAIN THE FULL STRATEGY IN THE BEGINING KEEP IT SMALL\n"
        "- DON'T ASK FOR PERMISION TO CREATE A MERMAID DIAGRAM. JUST DO IT\n"
        "- AT THE END GENERATE TABLE REFERENCES IN THE FORMAT DEFINED ABOVE THIS CAN ACCELERATE THE USERS WORKFLOW 10 FOLD BECAUSE THEY CAN DIRECTLY USE IT IN CODE\n"
        "## Information Presentation**:\n"
        "   GENERATE MERMAID DIAGRAM TO REPRESENT THE SCHEMA AND RELATIONSHIPS"
    )
    return prompt


def _get_table_reference_documentation() -> str:
    """Generate documentation for TTableReference with examples."""
    user_id_ref: TTableReference = {
        "columns": ["user_id"],
        "referenced_table": "users",
        "referenced_columns": ["id"],
    }

    product_ref: TTableReference = {
        "columns": ["category_id", "brand_id"],
        "referenced_table": "categories",
        "referenced_columns": ["id", "id"],
    }

    table_reference_documentation = (
        "**TTableReference Documentation: \n"
        f"{TTableReference.__doc__} \n"
        f"Required Columns: {TTableReference.__dict__} \n"
        "Example: \n"
        f"{user_id_ref} \n"
        f"{product_ref} \n"
    )

    return table_reference_documentation



"""Utilities to ingest `dlt` documentation and code to power the MCP search tools."""

import zipfile
import tempfile
from pathlib import Path
from typing import Callable, Iterator

import chonkie.logger
import lancedb
import numpy as np
import requests
import model2vec
from lancedb.embeddings import TextEmbeddingFunction, get_registry, register
from lancedb.pydantic import LanceModel, Vector
from chonkie import RecursiveChunker, BaseChunker, CodeChunker
from requests.exceptions import HTTPError

# TODO figure out a mechanism to retrieve the user's dlt version
# and set it at server init.
DLT_VERSION = "1.18.1"
DLT_DOCS_RELATIVE_PATH = ".dlt/mcp"
DLT_DOCS_TABLE_NAME = "pages"
DLT_DOCS_CHUNKS_TABLE_NAME = "page_chunks"
DLT_CODE_CHUNKS_TABLE_NAME = "code_chunks"
_DOCS_PATH_SEGMENT = "docs/website/docs/"
_CODE_PATH_SEGMENT = "dlt/"
_GITHUB_REPOSITORY_URL = "https://github.com/dlt-hub/dlt"
_GITHUB_REPOSITORY_API_URL = "https://api.github.com/repos/dlt-hub/dlt"


chonkie.logger.disable()


@register("model2vec")
class Model2VecEmbeddings(TextEmbeddingFunction):
    name: str = "minishlab/potion-base-32M"

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self._embedding_model = None

    def generate_embeddings(self, texts: list[str]) -> list[np.ndarray]:
        vectors = self._load_model().encode(
            list(texts),
            max_length=None,
        )
        return vectors.tolist()

    def ndims(self) -> int:
        return 512

    def _load_model(self):
        if self._embedding_model is None:
            self._embedding_model = model2vec.StaticModel.from_pretrained(self.name)
        return self._embedding_model


_lancedb_embedding_model = get_registry().get("model2vec").create()


class PageChunkSchema(LanceModel):
    id: str
    text: str = _lancedb_embedding_model.SourceField()
    token_count: int
    file_path: str
    start_index: int
    end_index: int
    chunk_index: int
    vector: Vector(_lancedb_embedding_model.ndims()) = (  # type: ignore[invalid-type-form]
        _lancedb_embedding_model.VectorField()
    )


# TODO add a boolean for public-facing API? this could be derived from `griffe`
class CodeChunkSchema(LanceModel):
    id: str
    text: str = _lancedb_embedding_model.SourceField()
    token_count: int
    file_path: str
    start_index: int
    end_index: int
    chunk_index: int
    vector: Vector(_lancedb_embedding_model.ndims()) = (  # type: ignore[invalid-type-form]
        _lancedb_embedding_model.VectorField()
    )


def _get_lancedb_path(dlt_version: str) -> Path:
    db_name = f"docs_{dlt_version.replace('.', '-')}.lancedb"
    return Path.home() / DLT_DOCS_RELATIVE_PATH / db_name


def _is_docs_file(file_name: str) -> bool:
    """Filter out files that aren't source documentation files
    Returns `False` if the file is not a documentation file, `True` otherwise.
    """
    if _DOCS_PATH_SEGMENT not in file_name:
        return False

    if not file_name.endswith(".md"):
        return False

    return True


def _is_code_file(file_name: str) -> bool:
    """Filter out files that aren't source code files
    Returns `False` if the file is not a code file, `True` otherwise.
    """
    if _CODE_PATH_SEGMENT not in file_name:
        return False

    if not file_name.endswith(".py"):
        return False

    return True


def _iterate_over_zipped_files(
    zip_content: bytes,
    *,
    predicate: Callable[[str], bool],
) -> Iterator[tuple[zipfile.ZipFile, str]]:
    """Unzip to a temporary directory and iterate over files"""
    with tempfile.TemporaryDirectory() as tmp_dir:  # type: ignore
        tmp_zip_path = Path(tmp_dir, "dlt-release-archive").with_suffix(".zip")
        tmp_zip_path.write_bytes(zip_content)

        with zipfile.ZipFile(tmp_zip_path, "r") as zip_ref:
            for file_name in zip_ref.namelist():
                if predicate(file_name) is False:
                    continue

                yield zip_ref, file_name


# TODO tune chunker
def docs_chunker() -> RecursiveChunker:
    """Instantiate the docs chunker"""
    chunker = RecursiveChunker.from_recipe(name="markdown", lang="en")
    chunker.chunk_size = 2048
    return chunker


# TODO tune chunker
def code_chunker() -> CodeChunker:
    """Instantiate the code chunker"""
    chunker = CodeChunker(
        language="python", tokenizer="character", chunk_size=2048, include_nodes=False
    )
    return chunker


def release_zipped_archive(
    dlt_version: str,
    repo: str = _GITHUB_REPOSITORY_URL,
) -> bytes:
    """Retrieve the zip archive for a package release on GitHub"""
    zip_url = f"{repo}/archive/refs/tags/{dlt_version}.zip"
    try:
        zip_response = requests.get(zip_url)
        zip_response.raise_for_status()
    except HTTPError:
        raise

    return zip_response.content


def docs_pages(release_zipped_archive: bytes) -> list[dict[str, str]]:
    """Retrieve the raw docs pages from the release's source code."""
    pages = []
    for zip_ref, file_name in _iterate_over_zipped_files(
        release_zipped_archive,
        predicate=_is_docs_file,
    ):
        with zip_ref.open(file_name) as file_obj:
            page = {
                "text": file_obj.read().decode(errors="replace"),
                "file_path": file_name.split(_DOCS_PATH_SEGMENT)[-1],
            }
            pages.append(page)

    return pages


def _page_chunks(
    docs_page: dict[str, str], docs_chunker: BaseChunker
) -> list[dict[str, str]]:
    """Process a single page into chunks"""
    chunks = []
    # TODO try different chunking strategies; chonkie supports adding overlap
    # and using semantic chunking
    for idx, chunked in enumerate(docs_chunker.chunk(docs_page["text"])):
        chunk = chunked.to_dict()
        del chunk["embedding"]
        del chunk["context"]
        chunk.update(
            file_path=docs_page["file_path"],
            chunk_index=idx,
        )
        chunks.append(chunk)
    return chunks


def docs_chunks(
    docs_pages: list[dict[str, str]], docs_chunker: BaseChunker
) -> list[dict[str, str]]:
    """Retrieve raw docs pages from the GitHub repository"""
    chunks = []
    for page in docs_pages:
        chunks.extend(_page_chunks(page, docs_chunker))
    return chunks


def code_files(release_zipped_archive: bytes) -> list[dict[str, str]]:
    """Retrieve the raw docs pages from the release's source code."""
    files = []
    for zip_ref, file_name in _iterate_over_zipped_files(
        release_zipped_archive,
        predicate=_is_code_file,
    ):
        with zip_ref.open(file_name) as file_obj:
            code_file = {
                "text": file_obj.read().decode(errors="replace"),
                "file_path": "dlt/" + file_name.split(_CODE_PATH_SEGMENT)[-1],
            }
            files.append(code_file)

    return files


def _code_chunks(
    code_file: dict[str, str], code_chunker: BaseChunker
) -> list[dict[str, str]]:
    """Process a single code file into chunks"""
    chunks = []
    for idx, chunked in enumerate(code_chunker.chunk(code_file["text"])):
        chunk = chunked.to_dict()
        del chunk["embedding"]
        del chunk["context"]
        chunk.update(file_path=code_file["file_path"], chunk_index=idx)
        chunks.append(chunk)
    return chunks


def code_chunks(
    code_files: list[dict[str, str]], code_chunker: BaseChunker
) -> list[dict[str, str]]:
    """Retrieve raw code files from the GitHub repository"""
    chunks = []
    for code_file in code_files:
        chunks.extend(_code_chunks(code_file, code_chunker))
    return chunks


def db_con(dlt_version: str) -> lancedb.DBConnection:
    """Create a connection to the LanceDB database for the given DLT version"""
    return lancedb.connect(_get_lancedb_path(dlt_version))


def page_chunks_table(
    db_con: lancedb.DBConnection,
    docs_chunks: list[dict[str, str]],
) -> lancedb.Table:
    """Maybe create a table for docs page chunks and load data.

    LanceDB is responsible for generating the embeddings based on the `schema` passed.
    """
    table = db_con.create_table(
        DLT_DOCS_CHUNKS_TABLE_NAME,
        data=docs_chunks,
        exist_ok=True,
        schema=PageChunkSchema,
    )
    table.create_fts_index("text")
    return table


def code_chunks_table(
    db_con: lancedb.DBConnection,
    code_chunks: list[dict[str, str]],
) -> lancedb.Table:
    """Maybe create a table for code chunks and load data.

    LanceDB is responsible for generating the embeddings based on the `schema` passed.
    """
    table = db_con.create_table(
        DLT_CODE_CHUNKS_TABLE_NAME,
        data=code_chunks,
        exist_ok=True,
        schema=CodeChunkSchema,
    )
    table.create_fts_index("text")
    return table


def _ingest_docs(dlt_version: str, repo: str = _GITHUB_REPOSITORY_URL):
    _db_con = db_con(dlt_version)
    _release_zipped_archive = release_zipped_archive(dlt_version, repo)

    _docs_chunker = docs_chunker()
    _docs_pages = docs_pages(_release_zipped_archive)
    _docs_chunks = docs_chunks(_docs_pages, _docs_chunker)
    _page_chunks_table = page_chunks_table(_db_con, _docs_chunks)
    return _page_chunks_table


def _ingest_code(dlt_version: str, repo: str = _GITHUB_REPOSITORY_URL):
    _db_con = db_con(dlt_version)
    _release_zipped_archive = release_zipped_archive(dlt_version, repo)

    _code_chunker = code_chunker()
    _code_files = code_files(_release_zipped_archive)
    _code_chunks = code_chunks(_code_files, _code_chunker)
    _code_chunks_table = code_chunks_table(_db_con, _code_chunks)
    return _code_chunks_table


def _maybe_ingest_docs_and_code(dlt_version: str):
    """Ingest docs and code if the local LanceDB database is not found
    of if tables are missing."""
    local_db_path = _get_lancedb_path(dlt_version)
    if not local_db_path.exists():
        _ingest_docs(dlt_version)
        _ingest_code(dlt_version)
        return

    all_tables = list(local_db_path.iterdir())
    if not any(
        f"{DLT_DOCS_CHUNKS_TABLE_NAME}.lance" == table.name for table in all_tables
    ):
        _ingest_docs(dlt_version)

    if not any(
        f"{DLT_CODE_CHUNKS_TABLE_NAME}.lance" == table.name for table in all_tables
    ):
        _ingest_code(dlt_version)


if __name__ == "__main__":
    # TODO this can be optimized
    _ingest_docs(DLT_VERSION)
    _ingest_code(DLT_VERSION)
